using System;
using System.Collections.Generic;

namespace org.khelekore.prtree {

  /** A builder of internal nodes used during bulk loading of a PR-Tree.
   *  A PR-Tree is build by building a pseudo R-Tree and grabbing the
   *  leaf nodes (and then repeating until you have just one root node).
   *  This class creates the leaf nodes without building the full pseudo tree.
   */
  internal class LeafBuilder {

    #region class ListData<>

    private class ListData<T> {
      // Same NodeUsage objects in both lists, just ordered differently.
      public List<NodeUsage<T>> sx;
      public List<NodeUsage<T>> sy;

      public ListData(List<NodeUsage<T>> sx, List<NodeUsage<T>> sy) {
        this.sx = sx;
        this.sy = sy;
      }

      public override String ToString() {
        return "{sx: " + sx + ", sy: " + sy + "}";
      }
    }

    #endregion // class ListData<>

    #region class NodeGetter<>

    private class NodeGetter<T> {

      /** 
       * @param data the lists to grab node data from
       * @param id the id of the nodes we may pick
       * @param size the number of nodes we may pick
       * @param xlow the lower start index for the x list 
       * @param ylow the lower start index for the x list
       * @param xhigh the upper start index for the x list
       * @param yhigh the upper start index for the x list
       */
      public NodeGetter(
        LeafBuilder builder,
        ListData<T> data, int id, int size,
        int xlow, int ylow, int xhigh, int yhigh
      ) {
        this.builder = builder;
        this.data = data;
        this.id = id;
        this.size = size;
        this.xlow = xlow;
        this.ylow = ylow;
        this.xhigh = xhigh;
        this.yhigh = yhigh;
      }

      public override String ToString() {
        return
          "{taken: " + taken +
          ", size: " + size + ", id: " + id + ", xlow: " + xlow +
          ", ylow: " + ylow + ", xhigh: " + xhigh + ", yhigh: " +
          yhigh + ", data: " + data + "}";
      }

      public bool HasMoreData {
        get { return ElementsLeft > 0; }
      }

      public int ElementsLeft {
        get { return size - taken; }
      }

      public T getFirstUnusedX() {
        taken++;
        NodeUsage<T> nu = getFirstUnusedXNode();
        nu.Use();
        data.sx[xlow - 1] = null;
        return nu.Data;
      }

      public T getFirstUnusedY() {
        taken++;
        while(ylow < yhigh && isUsedNode(data.sy, ylow))
          ylow++;
        NodeUsage<T> nu = data.sy[ylow];
        data.sy[ylow++] = null;
        nu.Use();
        return nu.Data;
      }

      public T getLastUnusedX() {
        taken++;
        NodeUsage<T> nu = getLastUnusedXNode();
        nu.Use();
        data.sx[xlow - 1] = null;
        return nu.Data;
      }

      public T getLastUnusedY() {
        taken++;
        while(yhigh > ylow && isUsedNode(data.sy, yhigh))
          yhigh--;
        NodeUsage<T> nu = data.sy[yhigh];
        data.sy[yhigh--] = null;
        nu.Use();
        return nu.Data;
      }

      /** Split the remaining data into two parts,
       *  one part with the low x values and one with the high x values.
       */
      public List<NodeGetter<T>> Split() {
        int e = ElementsLeft;
        int lowSize = (e + 1) / 2;
        int highSize = e - lowSize;
        int lowId = this.builder.listHolderId++;
        int highId = this.builder.listHolderId++;

        // save positions
        int xl = xlow;
        int xh = xhigh;

        // pick a low element to the low list, mark as low,
        // pick a high element to the high list, mark as high
        while(HasMoreData) {
          taken++;
          NodeUsage<T> nu = getFirstUnusedXNode();
          nu.User = lowId;
          if(HasMoreData) {
            taken++;
            nu = getLastUnusedXNode();
            nu.User = highId;
          }
        }

        NodeGetter<T> lhLow = new NodeGetter<T>(
          this.builder, data, lowId, lowSize, xl, ylow, xhigh, yhigh
        );
        NodeGetter<T> lhHigh = new NodeGetter<T>(
          this.builder, data, highId, highSize, xlow, ylow, xh, yhigh
        );
        List<NodeGetter<T>> ret = new List<NodeGetter<T>>(2);
        ret.Add(lhLow);
        ret.Add(lhHigh);
        return ret;
      }

      private bool isUsedNode(List<NodeUsage<T>> ls, int pos) {
        NodeUsage<T> nu = ls[pos];
        return nu == null || nu.IsUsed || nu.User != id;
      }

      private NodeUsage<T> getFirstUnusedXNode() {
        while(xlow < xhigh && isUsedNode(data.sx, xlow))
          xlow++;
        return data.sx[xlow++];
      }

      private NodeUsage<T> getLastUnusedXNode() {
        while(xhigh > xlow && isUsedNode(data.sx, xhigh))
          xhigh--;
        return data.sx[xhigh--];
      }

      private LeafBuilder builder;

      private ListData<T> data;
      private int taken = 0;
      private int size;
      private int id;

      // indexes for list scanning
      private int xlow;  // goes up as we pick low nodes
      private int ylow;
      private int xhigh; // goes down as we pick high nodes
      private int yhigh;

    }

    #endregion // class NodeGetter<>

    #region interface NodeFactory<>

    /** A factory that creates the nodes (either leaf or internal).
     */
    public interface NodeFactory<N> {
      /** Create a new node 
       * @param data the data entries for the node, fully filled.
       */
      N create(Object[] data);
    }

    #endregion // interface NodeFactory<>

    public LeafBuilder(int branchFactor) {
      this.branchFactor = branchFactor;
    }

    public void BuildLeafs<ItemType, LeafNodeType, ItemDerivedType>(
      ICollection<ItemDerivedType> ls,
      List<LeafNodeType> leafNodes,
      IComparer<ItemType> xSorter,
      IComparer<ItemType> ySorter,
      NodeFactory<LeafNodeType> nf
    ) where ItemDerivedType : ItemType {

      // To not waste so much memory we create two lists, sorted by xmin 
      // and ymin respectivly. The two lists hold the same objects so we
      // can modify the usage info from either list.
      listHolderId = 1;
      List<NodeUsage<ItemType>> lsx = new List<NodeUsage<ItemType>>(ls.Count);
      List<NodeUsage<ItemType>> lsy = new List<NodeUsage<ItemType>>(ls.Count);

      foreach(ItemType t in ls) {
        NodeUsage<ItemType> nu = new NodeUsage<ItemType>(t);
        lsx.Add(nu);
        lsy.Add(nu);
      }

      lsx.Sort(new NodeUsageSorter<ItemType>(xSorter));
      lsy.Sort(new NodeUsageSorter<ItemType>(ySorter));
      //Collections.sort(lsx, new NodeUsageSorter<T>(xSorter));
      //Collections.sort(lsy, new NodeUsageSorter<T>(ySorter));

      List<NodeGetter<ItemType>> toExpand = new List<NodeGetter<ItemType>>();
      ListData<ItemType> listData = new ListData<ItemType>(lsx, lsy);
      int top = lsx.Count - 1;
      toExpand.Add(
        new NodeGetter<ItemType>(this, listData, listHolderId++, lsx.Count, 0, 0, top, top)
      );

      internalBuildLeafs(toExpand, leafNodes, nf);
    }

    /** Construct the four edge nodes then split the rests of the nodes 
     *  in the middle and loop with the two middle sets.
     */
    private void internalBuildLeafs<T, N>(
      List<NodeGetter<T>> toExpand,
      List<N> leafNodes,
      NodeFactory<N> nf
    ) {
      while(toExpand.Count > 0) {
        NodeGetter<T> lh = toExpand[0];
        toExpand.RemoveAt(0);
        if(lh.HasMoreData)
          leafNodes.Add(getLowXNode(lh, nf));

        if(lh.HasMoreData)
          leafNodes.Add(getLowYNode(lh, nf));

        if(lh.HasMoreData)
          leafNodes.Add(getHighXNode(lh, nf));

        if(lh.HasMoreData)
          leafNodes.Add(getHighYNode(lh, nf));

        if(lh.HasMoreData) {
          List<NodeGetter<T>> splitted = lh.Split();
          toExpand.AddRange(splitted);
        }
      }
    }

    /** Get the size of the node we are about to create, 
     *  limit by elements left and by branchfactor
     */
    private int getNum<T>(NodeGetter<T> lh) {
      return Math.Min(lh.ElementsLeft, branchFactor);
    }

    private N getLowXNode<T, N>(NodeGetter<T> lh, NodeFactory<N> nf) {
      Object[] data = new Object[getNum(lh)];
      for(int i = 0; i < data.Length; i++)
        data[i] = lh.getFirstUnusedX();
      return nf.create(data);
    }

    private N getLowYNode<T, N>(NodeGetter<T> lh, NodeFactory<N> nf) {
      Object[] data = new Object[getNum(lh)];
      for(int i = 0; i < data.Length; i++)
        data[i] = lh.getFirstUnusedY();
      return nf.create(data);
    }

    private N getHighXNode<T, N>(NodeGetter<T> lh, NodeFactory<N> nf) {
      Object[] data = new Object[getNum(lh)];
      for(int i = 0; i < data.Length; i++)
        data[i] = lh.getLastUnusedX();
      return nf.create(data);
    }

    private N getHighYNode<T, N>(NodeGetter<T> lh, NodeFactory<N> nf) {
      Object[] data = new Object[getNum(lh)];
      for(int i = 0; i < data.Length; i++)
        data[i] = lh.getLastUnusedY();
      return nf.create(data);
    }

    private int branchFactor;
    private int listHolderId;

  }

} // namespace org.khelekore.prtree
